@@ -648,7 +648,7 @@ void TvCtrlPointAddDevice(
 {
 	char *deviceType = NULL;
 	char *friendlyName = NULL;
-	char presURL[200];
+	char *presURL = NULL;
 	char *baseURL = NULL;
 	char *relURL = NULL;
 	char *UDN = NULL;
@@ -676,7 +676,7 @@ void TvCtrlPointAddDevice(
 	baseURL = SampleUtil_GetFirstDocumentItem(DescDoc, "URLBase");
 	relURL = SampleUtil_GetFirstDocumentItem(DescDoc, "presentationURL");
 
-	ret = UpnpResolveURL((baseURL ? baseURL : location), relURL, presURL);
+	ret = UpnpResolveURL2((baseURL ? baseURL : location), relURL, &presURL);
 
 	if (UPNP_E_SUCCESS != ret)
 		SampleUtil_Print("Error generating presURL from %s + %s\n",
@@ -741,6 +741,10 @@ void TvCtrlPointAddDevice(
 			deviceNode->device.AdvrTimeOut = expires;
 			for (service = 0; service < TV_SERVICE_SERVCOUNT;
 			     service++) {
+				if (serviceId[service] == NULL) {
+					/* not found */
+					continue;
+				}
 				strcpy(deviceNode->device.TvService[service].
 				       ServiceId, serviceId[service]);
 				strcpy(deviceNode->device.TvService[service].
@@ -794,6 +798,8 @@ void TvCtrlPointAddDevice(
 		free(baseURL);
 	if (relURL)
 		free(relURL);
+	if (presURL)
+		free(presURL);
 	for (service = 0; service < TV_SERVICE_SERVCOUNT; service++) {
 		if (serviceId[service])
 			free(serviceId[service]);
@@ -998,30 +1004,34 @@ void TvCtrlPointHandleGetVar(
  *   Cookie -- Optional data specified during callback registration
  *
  ********************************************************************************/
-int TvCtrlPointCallbackEventHandler(Upnp_EventType EventType, void *Event, void *Cookie)
+int TvCtrlPointCallbackEventHandler(Upnp_EventType EventType, const void *Event, void *Cookie)
 {
-	/*int errCode = 0;*/
+	int errCode = 0;
 
 	SampleUtil_PrintEvent(EventType, Event);
 	switch ( EventType ) {
 	/* SSDP Stuff */
 	case UPNP_DISCOVERY_ADVERTISEMENT_ALIVE:
 	case UPNP_DISCOVERY_SEARCH_RESULT: {
-		struct Upnp_Discovery *d_event = (struct Upnp_Discovery *)Event;
+		const UpnpDiscovery *d_event = (UpnpDiscovery *)Event;
 		IXML_Document *DescDoc = NULL;
-		int ret;
+		const char *location = NULL;
+		int errCode = UpnpDiscovery_get_ErrCode(d_event);
 
-		if (d_event->ErrCode != UPNP_E_SUCCESS) {
-			SampleUtil_Print("Error in Discovery Callback -- %d\n",
-				d_event->ErrCode);
+		if (errCode != UPNP_E_SUCCESS) {
+			SampleUtil_Print(
+				"Error in Discovery Callback -- %d\n", errCode);
 		}
-		ret = UpnpDownloadXmlDoc(d_event->Location, &DescDoc);
-		if (ret != UPNP_E_SUCCESS) {
-			SampleUtil_Print("Error obtaining device description from %s -- error = %d\n",
-				d_event->Location, ret);
+
+		location = UpnpString_get_String(UpnpDiscovery_get_Location(d_event));
+		errCode = UpnpDownloadXmlDoc(location, &DescDoc);
+		if (errCode != UPNP_E_SUCCESS) {
+			SampleUtil_Print(
+				"Error obtaining device description from %s -- error = %d\n",
+				location, errCode);
 		} else {
 			TvCtrlPointAddDevice(
-				DescDoc, d_event->Location, d_event->Expires);
+				DescDoc, location, UpnpDiscovery_get_Expires(d_event));
 		}
 		if (DescDoc) {
 			ixmlDocument_free(DescDoc);
@@ -1033,90 +1043,92 @@ int TvCtrlPointCallbackEventHandler(Upnp_EventType EventType, void *Event, void
 		/* Nothing to do here... */
 		break;
 	case UPNP_DISCOVERY_ADVERTISEMENT_BYEBYE: {
-		struct Upnp_Discovery *d_event = (struct Upnp_Discovery *)Event;
+		UpnpDiscovery *d_event = (UpnpDiscovery *)Event;
+		int errCode = UpnpDiscovery_get_ErrCode(d_event);
+		const char *deviceId = UpnpString_get_String(
+		UpnpDiscovery_get_DeviceID(d_event));
 
-		if (d_event->ErrCode != UPNP_E_SUCCESS) {
-			SampleUtil_Print("Error in Discovery ByeBye Callback -- %d\n",
-					d_event->ErrCode);
+		if (errCode != UPNP_E_SUCCESS) {
+			SampleUtil_Print(
+				"Error in Discovery ByeBye Callback -- %d\n", errCode);
 		}
-		SampleUtil_Print("Received ByeBye for Device: %s\n", d_event->DeviceId);
-		TvCtrlPointRemoveDevice(d_event->DeviceId);
+		SampleUtil_Print("Received ByeBye for Device: %s\n", deviceId);
+		TvCtrlPointRemoveDevice(deviceId);
 		SampleUtil_Print("After byebye:\n");
 		TvCtrlPointPrintList();
 		break;
 	}
 	/* SOAP Stuff */
 	case UPNP_CONTROL_ACTION_COMPLETE: {
-		struct Upnp_Action_Complete *a_event = (struct Upnp_Action_Complete *)Event;
-
-		if (a_event->ErrCode != UPNP_E_SUCCESS) {
+		UpnpActionComplete *a_event = (UpnpActionComplete *)Event;
+		int errCode = UpnpActionComplete_get_ErrCode(a_event);
+		if (errCode != UPNP_E_SUCCESS) {
 			SampleUtil_Print("Error in  Action Complete Callback -- %d\n",
-					a_event->ErrCode);
+				errCode);
 		}
 		/* No need for any processing here, just print out results.
 		 * Service state table updates are handled by events. */
 		break;
 	}
 	case UPNP_CONTROL_GET_VAR_COMPLETE: {
-		struct Upnp_State_Var_Complete *sv_event = (struct Upnp_State_Var_Complete *)Event;
-
-		if (sv_event->ErrCode != UPNP_E_SUCCESS) {
-			SampleUtil_Print("Error in Get Var Complete Callback -- %d\n",
-					sv_event->ErrCode);
+		UpnpStateVarComplete *sv_event = (UpnpStateVarComplete *)Event;
+		int errCode = UpnpStateVarComplete_get_ErrCode(sv_event);
+		if (errCode != UPNP_E_SUCCESS) {
+			SampleUtil_Print(
+				"Error in Get Var Complete Callback -- %d\n", errCode);
 		} else {
 			TvCtrlPointHandleGetVar(
-				sv_event->CtrlUrl,
-				sv_event->StateVarName,
-				sv_event->CurrentVal);
+				UpnpString_get_String(UpnpStateVarComplete_get_CtrlUrl(sv_event)),
+				UpnpString_get_String(UpnpStateVarComplete_get_StateVarName(sv_event)),
+				UpnpStateVarComplete_get_CurrentVal(sv_event));
 		}
 		break;
 	}
 	/* GENA Stuff */
 	case UPNP_EVENT_RECEIVED: {
-		struct Upnp_Event *e_event = (struct Upnp_Event *)Event;
-
+		UpnpEvent *e_event = (UpnpEvent *)Event;
 		TvCtrlPointHandleEvent(
-			e_event->Sid,
-			e_event->EventKey,
-			e_event->ChangedVariables);
+			UpnpEvent_get_SID_cstr(e_event),
+			UpnpEvent_get_EventKey(e_event),
+			UpnpEvent_get_ChangedVariables(e_event));
 		break;
 	}
 	case UPNP_EVENT_SUBSCRIBE_COMPLETE:
 	case UPNP_EVENT_UNSUBSCRIBE_COMPLETE:
 	case UPNP_EVENT_RENEWAL_COMPLETE: {
-		struct Upnp_Event_Subscribe *es_event = (struct Upnp_Event_Subscribe *)Event;
+		UpnpEventSubscribe *es_event = (UpnpEventSubscribe *)Event;
 
-		if (es_event->ErrCode != UPNP_E_SUCCESS) {
-			SampleUtil_Print("Error in Event Subscribe Callback -- %d\n",
-					es_event->ErrCode);
+		errCode = UpnpEventSubscribe_get_ErrCode(es_event);
+		if (errCode != UPNP_E_SUCCESS) {
+			SampleUtil_Print(
+				"Error in Event Subscribe Callback -- %d\n", errCode);
 		} else {
 			TvCtrlPointHandleSubscribeUpdate(
-				es_event->PublisherUrl,
-				es_event->Sid,
-				es_event->TimeOut);
+				UpnpString_get_String(UpnpEventSubscribe_get_PublisherUrl(es_event)),
+				UpnpString_get_String(UpnpEventSubscribe_get_SID(es_event)),
+				UpnpEventSubscribe_get_TimeOut(es_event));
 		}
 		break;
 	}
 	case UPNP_EVENT_AUTORENEWAL_FAILED:
 	case UPNP_EVENT_SUBSCRIPTION_EXPIRED: {
-		struct Upnp_Event_Subscribe *es_event = (struct Upnp_Event_Subscribe *)Event;
+		UpnpEventSubscribe *es_event = (UpnpEventSubscribe *)Event;
 		int TimeOut = default_timeout;
 		Upnp_SID newSID;
-		int ret;
 
-		ret = UpnpSubscribe(
+		errCode = UpnpSubscribe(
 			ctrlpt_handle,
-			es_event->PublisherUrl,
+			UpnpString_get_String(UpnpEventSubscribe_get_PublisherUrl(es_event)),
 			&TimeOut,
 			newSID);
-		if (ret == UPNP_E_SUCCESS) {
+		if (errCode == UPNP_E_SUCCESS) {
 			SampleUtil_Print("Subscribed to EventURL with SID=%s\n", newSID);
 			TvCtrlPointHandleSubscribeUpdate(
-				es_event->PublisherUrl,
+				UpnpString_get_String(UpnpEventSubscribe_get_PublisherUrl(es_event)),
 				newSID,
 				TimeOut);
 		} else {
-			SampleUtil_Print("Error Subscribing to EventURL -- %d\n", ret);
+			SampleUtil_Print("Error Subscribing to EventURL -- %d\n", errCode);
 		}
 		break;
 	}
@@ -1219,9 +1231,9 @@ int TvCtrlPointStart(print_string printFunctionPtr, state_update updateFunctionP
 			 "\tipaddress = %s port = %u\n",
 			 ip_address ? ip_address : "{NULL}", port);
 
-	rc = UpnpInit(ip_address, port);
+	rc = UpnpInit2(ip_address, port);
 	if (rc != UPNP_E_SUCCESS) {
-		SampleUtil_Print("WinCEStart: UpnpInit() Error: %d\n", rc);
+		SampleUtil_Print("WinCEStart: UpnpInit2() Error: %d\n", rc);
 		if (!combo) {
 			UpnpFinish();
 
@@ -1451,7 +1463,9 @@ void *TvCtrlPointCommandLoop(void *args)
 
 	while (1) {
 		SampleUtil_Print("\n>> ");
-		fgets(cmdline, 100, stdin);
+		char *s = fgets(cmdline, 100, stdin);
+		if (!s)
+			break;
 		TvCtrlPointProcessCommand(cmdline);
 	}
 
