@@ -1,31 +1,31 @@
 /**************************************************************************
  *
- * Copyright (c) 2000-2003 Intel Corporation 
- * All rights reserved. 
- * Copyright (c) 2012 France Telecom All rights reserved. 
+ * Copyright (c) 2000-2003 Intel Corporation
+ * All rights reserved.
+ * Copyright (c) 2012 France Telecom All rights reserved.
  *
- * Redistribution and use in source and binary forms, with or without 
- * modification, are permitted provided that the following conditions are met: 
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions are met:
  *
- * - Redistributions of source code must retain the above copyright notice, 
- * this list of conditions and the following disclaimer. 
- * - Redistributions in binary form must reproduce the above copyright notice, 
- * this list of conditions and the following disclaimer in the documentation 
- * and/or other materials provided with the distribution. 
- * - Neither name of Intel Corporation nor the names of its contributors 
- * may be used to endorse or promote products derived from this software 
+ * - Redistributions of source code must retain the above copyright notice,
+ * this list of conditions and the following disclaimer.
+ * - Redistributions in binary form must reproduce the above copyright notice,
+ * this list of conditions and the following disclaimer in the documentation
+ * and/or other materials provided with the distribution.
+ * - Neither name of Intel Corporation nor the names of its contributors
+ * may be used to endorse or promote products derived from this software
  * without specific prior written permission.
- * 
- * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
- * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
- * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
- * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL INTEL OR 
- * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
- * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
- * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
- * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY 
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
+ * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL INTEL OR
+ * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+ * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+ * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
+ * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
  * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
- * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
+ * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
  **************************************************************************/
@@ -43,6 +43,9 @@
 
 #include "webserver.h"
 
+
+#include "ExtraHeaders.h"
+#include "FileInfo.h"
 #include "httpparser.h"
 #include "httpreadwrite.h"
 #include "ithread.h"
@@ -62,7 +65,7 @@
 #include <fcntl.h>
 #include <sys/stat.h>
 
-#ifdef WIN32
+#ifdef _WIN32
 	 #define snprintf _snprintf
 #endif
 
@@ -133,7 +136,7 @@ static const char *gMediaTypes[] = {
 #define NUM_HTTP_HEADER_NAMES 33
 
 #define ASCTIME_R_BUFFER_SIZE 26
-#ifdef WIN32
+#ifdef _WIN32
 static char *web_server_asctime_r(const struct tm *tm, char *buf)
 {
 	if (tm == NULL || buf == NULL)
@@ -313,7 +316,7 @@ static UPNP_INLINE int get_content_type(
 	/*! [in] . */
 	const char *filename,
 	/*! [out] . */
-	DOMString *content_type)
+	OUT UpnpFileInfo *fileInfo)
 {
 	const char *extension;
 	const char *type;
@@ -323,7 +326,7 @@ static UPNP_INLINE int get_content_type(
 	size_t length = 0;
 	int rc = 0;
 
-	(*content_type) = NULL;
+	UpnpFileInfo_set_ContentType(fileInfo, NULL);
 	/* get ext */
 	extension = strrchr(filename, '.');
 	if (extension != NULL)
@@ -343,9 +346,9 @@ static UPNP_INLINE int get_content_type(
 		free(temp);
 		return UPNP_E_OUTOF_MEMORY;
 	}
-	(*content_type) = ixmlCloneDOMString(temp);
+	UpnpFileInfo_set_ContentType(fileInfo, temp);
 	free(temp);
-	if (!(*content_type))
+	if (!UpnpFileInfo_get_ContentType(fileInfo))
 		return UPNP_E_OUTOF_MEMORY;
 
 	return 0;
@@ -416,7 +419,7 @@ static void alias_release(
 	ithread_mutex_unlock(&gWebMutex);
 }
 
-int web_server_set_alias(const char *alias_name, 
+int web_server_set_alias(const char *alias_name,
 	const char *alias_content, size_t alias_content_length,
 	time_t last_modified)
 {
@@ -518,39 +521,41 @@ static int get_file_info(
 	/*! [out] File information object having file attributes such as filelength,
 	 * when was the file last modified, whether a file or a directory and
 	 * whether the file or directory is readable. */
-	struct File_Info *info)
+	OUT UpnpFileInfo *info)
 {
 	int code;
 	struct stat s;
 	FILE *fp;
 	int rc = 0;
+	time_t aux_LastModified;
 	struct tm date;
 	char buffer[ASCTIME_R_BUFFER_SIZE];
 
-	ixmlFreeDOMString(info->content_type);	
-	info->content_type = NULL;
+	UpnpFileInfo_set_ContentType(info, NULL);
 	code = stat(filename, &s);
 	if (code == -1)
 		return -1;
 	if (S_ISDIR(s.st_mode))
-		info->is_directory = TRUE;
+		UpnpFileInfo_set_IsDirectory(info, TRUE);
 	else if (S_ISREG(s.st_mode))
-		info->is_directory = FALSE;
+		UpnpFileInfo_set_IsDirectory(info, FALSE);
 	else
 		return -1;
 	/* check readable */
 	fp = fopen(filename, "r");
-	info->is_readable = (fp != NULL);
+	UpnpFileInfo_set_IsReadable(info, fp != NULL);
 	if (fp)
 		fclose(fp);
-	info->file_length = s.st_size;
-	info->last_modified = s.st_mtime;
-	rc = get_content_type(filename, &info->content_type);
+	UpnpFileInfo_set_FileLength(info, s.st_size);
+	UpnpFileInfo_set_LastModified(info, s.st_mtime);
+	rc = get_content_type(filename, info);
+	aux_LastModified = UpnpFileInfo_get_LastModified(info);
 	UpnpPrintf(UPNP_INFO, HTTP, __FILE__, __LINE__,
 		"file info: %s, length: %lld, last_mod=%s readable=%d\n",
-		filename, (long long)info->file_length,
-		web_server_asctime_r(http_gmtime_r(&info->last_modified, &date), buffer),
-		info->is_readable);
+		filename,
+		(long long)UpnpFileInfo_get_FileLength(info),
+		web_server_asctime_r(http_gmtime_r(&aux_LastModified, &date), buffer),
+		UpnpFileInfo_get_IsReadable(info));
 
 	return rc;
 }
@@ -588,15 +593,14 @@ static UPNP_INLINE int get_alias(
 	struct xml_alias_t *alias,
 	/*! [out] File information object which will be filled up if the file
 	 * comparison succeeds. */
-	struct File_Info *info)
+	UpnpFileInfo *info)
 {
 	int cmp = strcmp(alias->name.buf, request_file);
 	if (cmp == 0) {
-		/* fill up info */
-		info->file_length = (off_t)alias->doc.length;
-		info->is_readable = TRUE;
-		info->is_directory = FALSE;
-		info->last_modified = alias->last_modified;
+		UpnpFileInfo_set_FileLength(info, (off_t)alias->doc.length);
+		UpnpFileInfo_set_IsDirectory(info, FALSE);
+		UpnpFileInfo_set_IsReadable(info, TRUE);
+		UpnpFileInfo_set_LastModified(info, alias->last_modified);
 	}
 
 	return cmp == 0;
@@ -610,7 +614,9 @@ static UPNP_INLINE int get_alias(
  */
 static int isFileInVirtualDir(
 	/*! [in] Directory path to be tested for virtual directory. */
-	char *filePath)
+	char *filePath,
+	/*! [out] The cookie registered with this virtual directory, if matched. */
+	const void **cookie)
 {
 	virtualDirList *pCurVirtualDir;
 	size_t webDirLen;
@@ -621,15 +627,21 @@ static int isFileInVirtualDir(
 		if (webDirLen) {
 			if (pCurVirtualDir->dirName[webDirLen - 1] == '/') {
 				if (strncmp(pCurVirtualDir->dirName, filePath,
-						webDirLen) == 0)
+						webDirLen) == 0) {
+					if (cookie != NULL)
+						*cookie = pCurVirtualDir->cookie;
 					return !0;
+				}
 			} else {
 				if (strncmp(pCurVirtualDir->dirName, filePath,
 						webDirLen) == 0 &&
 				    (filePath[webDirLen] == '/' ||
 				     filePath[webDirLen] == '\0' ||
-				     filePath[webDirLen] == '?'))
+				     filePath[webDirLen] == '?')) {
+					if (cookie != NULL)
+						*cookie = pCurVirtualDir->cookie;
 					return !0;
+				}
 			}
 		}
 		pCurVirtualDir = pCurVirtualDir->next;
@@ -1008,81 +1020,62 @@ static int CheckOtherHTTPHeaders(
 	return RetCode;
 }
 
+static void FreeExtraHTTPHeaders(
+	/*! [in] extra HTTP headers to free. */
+	const struct list_head *extraHeadersList)
+{
+	struct list_head *pos;
+	struct list_head *tmp;
+	UpnpExtraHeaders *extra;
+
+	list_for_each_safe(pos, tmp, extraHeadersList) {
+		extra = (UpnpExtraHeaders *)pos;
+		UpnpExtraHeaders_delete(extra);
+	}
+}
+
 /*!
  * \brief Build an array of unrecognized headers.
  *
  * \return nothing
  */
-#define MAX_EXTRA_HEADERS 128
 static int ExtraHTTPHeaders(
 	/*! [in] HTTP Request message. */
 	http_message_t *Req,
-	struct Extra_Headers **ExtraHeaders)
+	const struct list_head *extraHeadersList)
 {
 	http_header_t *header;
 	ListNode *node;
-	int index, nb_extra = 0;
-	struct Extra_Headers *extra_headers;
+	int index;
+	UpnpExtraHeaders *extraHeader;
+	struct list_head *extraHeaderNode;
 
 	node = ListHead(&Req->headers);
-	extra_headers = *ExtraHeaders =
-		(struct Extra_Headers*) malloc(MAX_EXTRA_HEADERS * sizeof(struct Extra_Headers));
-	if (!extra_headers) {
-		return HTTP_INTERNAL_SERVER_ERROR;
-	}
 	while (node != NULL) {
-		header = (http_header_t *) node->item;
+		header = (http_header_t *)node->item;
 		/* find header type. */
 		index = map_str_to_int((const char *)header->name.buf,
 				header->name.length, Http_Header_Names,
 				NUM_HTTP_HEADER_NAMES, FALSE);
 		if (index < 0) {
-			extra_headers->name = (char *)malloc(header->name.length + 1);
-			extra_headers->value = (char *)malloc(header->value.length + 1);
-			if (!extra_headers->name || !extra_headers->value) {
-				/* cleanup will be made by caller */
+			extraHeader = UpnpExtraHeaders_new();
+			if (!extraHeader) {
+				FreeExtraHTTPHeaders(extraHeadersList);
 				return HTTP_INTERNAL_SERVER_ERROR;
 			}
-			memcpy(extra_headers->name, header->name.buf, header->name.length);
-			memcpy(extra_headers->value, header->value.buf, header->value.length);
-			extra_headers->name[header->name.length] = '\0';
-			extra_headers->value[header->value.length] = '\0';
-			extra_headers->resp = NULL;
-
-			extra_headers++;
-			nb_extra++;
-
-			if (nb_extra == MAX_EXTRA_HEADERS - 1) {
-				break;
-			}
+			/* TODO: Check that cast that removes const. */
+			/*extraHeaderNode = (struct list_head *)ExtraHeaders_get_node(extraHeader);*/
+			extraHeaderNode = UpnpExtraHeaders_get_node(extraHeader);
+			list_add(extraHeaderNode, extraHeadersList);
+			UpnpExtraHeaders_strncpy_name(extraHeader, header->name.buf, header->name.length);
+			UpnpExtraHeaders_strncpy_value(extraHeader, header->value.buf, header->value.length);
 		}
 		node = ListNext(&Req->headers, node);
 	}
-	extra_headers->name = extra_headers->value = extra_headers->resp = NULL;
+
 	return HTTP_OK;
 }
 
-static void FreeExtraHTTPHeaders(
-	/*! [in] extra HTTP headers to free. */
-	struct Extra_Headers *ExtraHeaders)
-{
-	struct Extra_Headers *extra_headers = ExtraHeaders;
-
-	if (!ExtraHeaders) {
-		return;
-	}
-
-	while (extra_headers->name) {
-		free(extra_headers->name);
-		if (extra_headers->value) free(extra_headers->value);
-		if (extra_headers->resp) ixmlFreeDOMString(extra_headers->resp);
-		extra_headers++;
-	}
-
-	free(ExtraHeaders);
-}
-
-
 /*!
  * \brief Processes the request and returns the result in the output parameters.
  *
@@ -1113,7 +1106,8 @@ static int process_request(
 	int err_code;
 
 	char *request_doc;
-	struct File_Info finfo;
+	UpnpFileInfo *finfo;
+	time_t aux_LastModified;
 	int using_alias;
 	int using_virtual_dir;
 	uri_type *url;
@@ -1122,7 +1116,6 @@ static int process_request(
 	int resp_minor;
 	int alias_grabbed;
 	size_t dummy;
-	struct Extra_Headers *extra_headers = NULL;
 
 	print_http_headers(req);
 	url = &req->uri;
@@ -1133,7 +1126,7 @@ static int process_request(
 	/* init */
 	memset(&finfo, 0, sizeof(finfo));
 	request_doc = NULL;
-	finfo.content_type = NULL;
+	finfo = UpnpFileInfo_new();
 	alias_grabbed = FALSE;
 	err_code = HTTP_INTERNAL_SERVER_ERROR;	/* default error */
 	using_virtual_dir = FALSE;
@@ -1162,7 +1155,7 @@ static int process_request(
 		err_code = HTTP_BAD_REQUEST;
 		goto error_handler;
 	}
-	if (isFileInVirtualDir(request_doc)) {
+	if (isFileInVirtualDir(request_doc, &RespInstr->Cookie)) {
 		using_virtual_dir = TRUE;
 		RespInstr->IsVirtualFile = 1;
 		if (membuffer_assign_str(filename, request_doc) != 0) {
@@ -1173,12 +1166,11 @@ static int process_request(
 		if (is_valid_alias(&gAliasDoc)) {
 			alias_grab(alias);
 			alias_grabbed = TRUE;
-			using_alias = get_alias(request_doc, alias, &finfo);
+			using_alias = get_alias(request_doc, alias, finfo);
 			if (using_alias == TRUE) {
-				finfo.content_type =
-				    ixmlCloneDOMString("text/xml");
-
-				if (finfo.content_type == NULL) {
+				UpnpFileInfo_set_ContentType(finfo,
+					"text/xml; charset=\"utf-8\"");
+				if (UpnpFileInfo_get_ContentType(finfo) == NULL) {
 					goto error_handler;
 				}
 			}
@@ -1186,39 +1178,36 @@ static int process_request(
 	}
 	if (using_virtual_dir) {
 		if (req->method != HTTPMETHOD_POST) {
-			if ((code = ExtraHTTPHeaders(req, &extra_headers)) != HTTP_OK) {
+			if ((code = ExtraHTTPHeaders(req, UpnpFileInfo_get_ExtraHeadersList(finfo))) != HTTP_OK) {
 				err_code = code;
 				goto error_handler;
 			}
+
 			/* get file info */
-			finfo.extra_headers = extra_headers;
 			if (virtualDirCallback.
-			    get_info(filename->buf, &finfo) != 0) {
+			    get_info(filename->buf, finfo, RespInstr->Cookie, &RespInstr->RequestCookie) != 0) {
 				err_code = HTTP_NOT_FOUND;
 				goto error_handler;
 			}
 			/* try index.html if req is a dir */
-			if (finfo.is_directory) {
+			if (UpnpFileInfo_get_IsDirectory(finfo)) {
 				if (filename->buf[filename->length - 1] == '/') {
 					temp_str = "index.html";
 				} else {
 					temp_str = "/index.html";
 				}
-				if (membuffer_append_str(filename, temp_str) !=
-				    0) {
+				if (membuffer_append_str(filename, temp_str) != 0) {
 					goto error_handler;
 				}
 				/* get info */
-				if ((virtualDirCallback.
-				     get_info(filename->buf,
-					      &finfo) != UPNP_E_SUCCESS)
-				    || finfo.is_directory) {
+				if (virtualDirCallback.get_info(filename->buf, finfo, RespInstr->Cookie, &RespInstr->RequestCookie) != UPNP_E_SUCCESS ||
+				    UpnpFileInfo_get_IsDirectory(finfo)) {
 					err_code = HTTP_NOT_FOUND;
 					goto error_handler;
 				}
 			}
 			/* not readable */
-			if (!finfo.is_readable) {
+			if (!UpnpFileInfo_get_IsReadable(finfo)) {
 				err_code = HTTP_FORBIDDEN;
 				goto error_handler;
 			}
@@ -1248,12 +1237,12 @@ static int process_request(
 		}
 		if (req->method != HTTPMETHOD_POST) {
 			/* get info on file */
-			if (get_file_info(filename->buf, &finfo) != 0) {
+			if (get_file_info(filename->buf, finfo) != 0) {
 				err_code = HTTP_NOT_FOUND;
 				goto error_handler;
 			}
 			/* try index.html if req is a dir */
-			if (finfo.is_directory) {
+			if (UpnpFileInfo_get_IsDirectory(finfo)) {
 				if (filename->buf[filename->length - 1] == '/') {
 					temp_str = "index.html";
 				} else {
@@ -1264,14 +1253,14 @@ static int process_request(
 					goto error_handler;
 				}
 				/* get info */
-				if (get_file_info(filename->buf, &finfo) != 0 ||
-				    finfo.is_directory) {
+				if (get_file_info(filename->buf, finfo) != 0 ||
+				    UpnpFileInfo_get_IsDirectory(finfo)) {
 					err_code = HTTP_NOT_FOUND;
 					goto error_handler;
 				}
 			}
 			/* not readable */
-			if (!finfo.is_readable) {
+			if (!UpnpFileInfo_get_IsReadable(finfo)) {
 				err_code = HTTP_FORBIDDEN;
 				goto error_handler;
 			}
@@ -1282,15 +1271,14 @@ static int process_request(
 		/*          goto error_handler; */
 		/*      } */
 	}
-	RespInstr->ReadSendSize = finfo.file_length;
+	RespInstr->ReadSendSize = UpnpFileInfo_get_FileLength(finfo);
 	/* Check other header field. */
-	if ((code =
-		 CheckOtherHTTPHeaders(req, RespInstr,
-				   finfo.file_length)) != HTTP_OK) {
+	code = CheckOtherHTTPHeaders(req, RespInstr,
+		UpnpFileInfo_get_FileLength(finfo));
+	if (code != HTTP_OK) {
 		err_code = code;
 		goto error_handler;
 	}
-
 	if (req->method == HTTPMETHOD_POST) {
 		*rtype = RESP_POST;
 		err_code = HTTP_OK;
@@ -1298,7 +1286,7 @@ static int process_request(
 	}
 
 	/* Check if chunked encoding should be used. */
-	if (using_virtual_dir && finfo.file_length == UPNP_USING_CHUNKED) {
+	if (using_virtual_dir && UpnpFileInfo_get_FileLength(finfo) == UPNP_USING_CHUNKED) {
 		/* Chunked encoding is only supported by HTTP 1.1 clients */
 		if (resp_major == 1 && resp_minor == 1) {
 			RespInstr->IsChunkActive = 1;
@@ -1311,71 +1299,75 @@ static int process_request(
 		}
 	}
 
+	aux_LastModified = UpnpFileInfo_get_LastModified(finfo);
 	if (RespInstr->IsRangeActive && RespInstr->IsChunkActive) {
 		/* Content-Range: bytes 222-3333/4000  HTTP_PARTIAL_CONTENT */
 		/* Transfer-Encoding: chunked */
 		if (http_MakeMessage(headers, resp_major, resp_minor,
 		    "R" "T" "GKLD" "s" "tcS" "Xc" "ECc",
 		    HTTP_PARTIAL_CONTENT,	/* status code */
-		    finfo.content_type,	/* content type */
+		    UpnpFileInfo_get_ContentType(finfo), /* content type */
 		    RespInstr,	/* range info */
 		    RespInstr,	/* language info */
 		    "LAST-MODIFIED: ",
-		    &finfo.last_modified,
-		    X_USER_AGENT, extra_headers) != 0) {
+		    &aux_LastModified,
+		    X_USER_AGENT,
+		    UpnpFileInfo_get_ExtraHeadersList(finfo)) != 0) {
 			goto error_handler;
 		}
 	} else if (RespInstr->IsRangeActive && !RespInstr->IsChunkActive) {
 		/* Content-Range: bytes 222-3333/4000  HTTP_PARTIAL_CONTENT */
 		if (http_MakeMessage(headers, resp_major, resp_minor,
-			"R" "N" "T" "GLD" "s" "tcS" "Xc" "ECc",
-			HTTP_PARTIAL_CONTENT,	/* status code */
-			RespInstr->ReadSendSize,	/* content length */
-			finfo.content_type,	/* content type */
-			RespInstr,	/* range info */
-			RespInstr,	/* language info */
-			"LAST-MODIFIED: ",
-			&finfo.last_modified,
-			X_USER_AGENT, extra_headers) != 0) {
+		    "R" "N" "T" "GLD" "s" "tcS" "Xc" "ECc",
+		    HTTP_PARTIAL_CONTENT,	/* status code */
+		    RespInstr->ReadSendSize,	/* content length */
+		    UpnpFileInfo_get_ContentType(finfo), /* content type */
+		    RespInstr,	/* range info */
+		    RespInstr,	/* language info */
+		    "LAST-MODIFIED: ",
+		    &aux_LastModified,
+		    X_USER_AGENT,
+		    UpnpFileInfo_get_ExtraHeadersList(finfo)) != 0) {
 			goto error_handler;
 		}
 	} else if (!RespInstr->IsRangeActive && RespInstr->IsChunkActive) {
 		/* Transfer-Encoding: chunked */
 		if (http_MakeMessage(headers, resp_major, resp_minor,
-			"RK" "TLD" "s" "tcS" "Xc" "ECc",
-			HTTP_OK,	/* status code */
-			finfo.content_type,	/* content type */
-			RespInstr,	/* language info */
-			"LAST-MODIFIED: ",
-			&finfo.last_modified,
-			X_USER_AGENT, extra_headers) != 0) {
+		    "RK" "TLD" "s" "tcS" "Xc" "ECc",
+		    HTTP_OK,	/* status code */
+		    UpnpFileInfo_get_ContentType(finfo), /* content type */
+		    RespInstr,	/* language info */
+		    "LAST-MODIFIED: ",
+		    &aux_LastModified,
+		    X_USER_AGENT,
+		    UpnpFileInfo_get_ExtraHeadersList(finfo)) != 0) {
 			goto error_handler;
 		}
 	} else {
 		/* !RespInstr->IsRangeActive && !RespInstr->IsChunkActive */
 		if (RespInstr->ReadSendSize >= 0) {
 			if (http_MakeMessage(headers, resp_major, resp_minor,
-				"R" "N" "TLD" "s" "tcS" "Xc" "ECc",
-				HTTP_OK,	/* status code */
-				RespInstr->ReadSendSize,	/* content length */
-				finfo.content_type,	/* content type */
-				RespInstr,	/* language info */
-				"LAST-MODIFIED: ",
-				&finfo.last_modified,
-				X_USER_AGENT,
-				extra_headers) != 0) {
+			    "R" "N" "TLD" "s" "tcS" "Xc" "ECc",
+			    HTTP_OK,	/* status code */
+			    RespInstr->ReadSendSize,	/* content length */
+			    UpnpFileInfo_get_ContentType(finfo), /* content type */
+			    RespInstr,	/* language info */
+			    "LAST-MODIFIED: ",
+			    &aux_LastModified,
+			    X_USER_AGENT,
+			    UpnpFileInfo_get_ExtraHeadersList(finfo)) != 0) {
 				goto error_handler;
 			}
 		} else {
 			if (http_MakeMessage(headers, resp_major, resp_minor,
-				"R" "TLD" "s" "tcS" "Xc" "ECc",
-				HTTP_OK,	/* status code */
-				finfo.content_type,	/* content type */
-				RespInstr,	/* language info */
-				"LAST-MODIFIED: ",
-				&finfo.last_modified,
-				X_USER_AGENT,
-				extra_headers) != 0) {
+			    "R" "TLD" "s" "tcS" "Xc" "ECc",
+			    HTTP_OK,	/* status code */
+			    UpnpFileInfo_get_ContentType(finfo), /* content type */
+			    RespInstr,	/* language info */
+			    "LAST-MODIFIED: ",
+			    &aux_LastModified,
+			    X_USER_AGENT,
+			    UpnpFileInfo_get_ExtraHeadersList(finfo)) != 0) {
 				goto error_handler;
 			}
 		}
@@ -1400,8 +1392,8 @@ static int process_request(
 
  error_handler:
 	free(request_doc);
-	FreeExtraHTTPHeaders(extra_headers);
-	ixmlFreeDOMString(finfo.content_type);
+	FreeExtraHTTPHeaders(UpnpFileInfo_get_ExtraHeadersList(finfo));
+	UpnpFileInfo_delete(finfo);
 	if (err_code != HTTP_OK && alias_grabbed) {
 		alias_release(alias);
 	}
@@ -1432,7 +1424,7 @@ static int http_RecvPostMessage(
 {
 	size_t Data_Buf_Size = 1024;
 	char Buf[1024];
-	int Timeout = 0;
+	int Timeout = -1;
 	FILE *Fp;
 	parse_status_t status = PARSE_OK;
 	int ok_on_close = FALSE;
@@ -1441,7 +1433,7 @@ static int http_RecvPostMessage(
 	int ret_code = HTTP_OK;
 
 	if (Instr && Instr->IsVirtualFile) {
-		Fp = (virtualDirCallback.open) (filename, UPNP_WRITE);
+		Fp = (virtualDirCallback.open) (filename, UPNP_WRITE, Instr->Cookie, Instr->RequestCookie);
 		if (Fp == NULL)
 			return HTTP_INTERNAL_SERVER_ERROR;
 	} else {
@@ -1519,7 +1511,8 @@ static int http_RecvPostMessage(
 		       Data_Buf_Size);
 		entity_offset += Data_Buf_Size;
 		if (Instr && Instr->IsVirtualFile) {
-			int n = virtualDirCallback.write(Fp, Buf, Data_Buf_Size);
+			int n = virtualDirCallback.write(Fp, Buf, Data_Buf_Size,
+					Instr->Cookie, Instr->RequestCookie);
 			if (n < 0) {
 				ret_code = HTTP_INTERNAL_SERVER_ERROR;
 				goto ExitFunction;
@@ -1535,7 +1528,7 @@ static int http_RecvPostMessage(
 		 entity_offset != parser->msg.entity.length);
 ExitFunction:
 	if (Instr && Instr->IsVirtualFile) {
-		virtualDirCallback.close(Fp);
+		virtualDirCallback.close(Fp, Instr->Cookie, Instr->RequestCookie);
 	} else {
 		fclose(Fp);
 	}
@@ -1547,21 +1540,15 @@ void web_server_callback(http_parser_t *parser, INOUT http_message_t *req,
 	SOCKINFO *info)
 {
 	int ret;
-	int timeout = 0;
+	int timeout = -1;
 	enum resp_type rtype = 0;
 	membuffer headers;
 	membuffer filename;
 	struct xml_alias_t xmldoc;
 	struct SendInstruction RespInstr;
 
-	/*Initialize instruction header. */
-	RespInstr.IsVirtualFile = 0;
-	RespInstr.IsChunkActive = 0;
-	RespInstr.IsRangeActive = 0;
-	RespInstr.IsTrailers = 0;
-	memset(RespInstr.AcceptLanguageHeader, 0,
-	       sizeof(RespInstr.AcceptLanguageHeader));
 	/* init */
+	memset(&RespInstr, 0, sizeof(RespInstr));
 	membuffer_init(&headers);
 	membuffer_init(&filename);
 
